#include "SliceEditor.hh"
#include "FileSystem.hh"
#include <cstdio>

SliceEditor::SliceEditor() : RequestManager("SliceEditor")
{
  m_log.debugL("Contructed SliceEditor");
  m_log.setMessageLevel(MESSAGE_LEVEL_DEBUG_MEDIUM);

  m_newIndex = 0;
  m_newGindex = 0;

  m_ignoreLoaded = false;

  ConfigSection * depth;
  if (m_tf->getConfig() && (depth = m_tf->getConfig()->findSection("Depth")))
  {
    loadConfigFile(depth);
  }
}

int SliceEditor::m_verifyIndex(0);
int SliceEditor::m_verifyEmpty(0);
int SliceEditor::m_editFilterLevel(0);
int SliceEditor::m_createExtendedTcov(0);
string SliceEditor::m_outfilePrefix("");
ofstream * SliceEditor::m_editTcov(NULL);
ofstream * SliceEditor::m_origTcov(NULL);

SliceEditor::~SliceEditor()
{
  m_log.debugL("Destructed SliceEditor");
}

void SliceEditor::loadConfigFile(ConfigSection * depth)
{
  m_ignoreLoaded = true;
  ConfigSection * id;
  while ((id = depth->getNextSection()))
  {
    string contig_id = id->getSectionName();
    string var;

    while ((var = id->getNextVariable()) != "")
    {
      string val = id->getValue(var);

      string::size_type equalpos = val.find(',');
      if (equalpos != val.npos)
      {
        string startpos = val.substr(0, equalpos);
        string endpos   = val.substr(equalpos+1, val.length()-equalpos);

        long start = atol(startpos.c_str());
        long end = atol(endpos.c_str());

        cerr << "Ignore from " << contig_id 
             << ":" << start << "->" << end << endl;

        m_ignoreMap[start] = end;
      }
    }
  }
}

bool SliceEditor::doEditPosition(long gindex)
{
  if (!m_ignoreLoaded)
  {
    return true;
  }

  bool retval = 0;
  gindex++;

  IgnorePosMap::const_iterator i = m_ignoreMap.upper_bound(gindex);
  i--;

  if (i != m_ignoreMap.end())
  {
    retval = !(i->first <= gindex && gindex <= i->second);

    if (!retval)
    {
      cerr << "Ignore position " << gindex << endl;
    }
  }

  return retval;
}

void SliceEditor::addSliceRange(long & index, long & gindex)
{
  m_out << "  </ReadCollection>" << endl;

  m_out << "  <SliceRange Index=\"" << index
        << "\" Gindex=\""           << gindex
        << "\">" << endl;

  index = gindex = 0L;
}

void SliceEditor::addSlice(SliceData * slice)
{
  if (m_doLogging) { m_log << slice; }

  if (m_createExtendedTcov) { writeExtendedTcov(slice, origtype); }

  long origGindex = slice->getGindex();
  char consensus  = slice->getConsensus();

  unsigned int sliceDepth = slice->getDepth();
 
  // Update each seqpos
  if (m_verifyIndex)
  {
    for (unsigned int i = 0; i < sliceDepth; i++)
    {
      if (slice->getBase(i) != '-')
      {
        const ReadData * rd = slice->getRead(i);
        const string & seqname = rd->getSeqname();

        if (rd->isReverseCompliment()) { m_seqPos[seqname]--; }
        else                           { m_seqPos[seqname]++; }

      }
    }
  }

  // Apply all edits at this location
  IndexEditMap::iterator edit;
  unsigned int removeCount = 0;

  for (edit =  m_indexEdits.find(origGindex);
       edit != m_indexEdits.end() && edit->first == origGindex;
       edit++)
  {
    if (edit->second)
    {
      if (!sliceDepth)
      {
        cerr << "Edit requested for dcov=0 slice at " << origGindex << endl;
      }

      edit->second->setVerifyIndex(m_verifyIndex);

      // Get the seqpos of the edited seq
      long seqpos = 0;
      if (m_verifyIndex) { seqpos = m_seqPos[edit->second->getSeqname()]; }

      // Apply the edit
      removeCount += edit->second->apply(slice, seqpos);

      //Could pull the edit out of the map, but really isn't necessary
      delete edit->second;
      edit->second = NULL;
    }
  }


  // Recall consensus after editing
  if (m_doConsensus 
      && doEditPosition(origGindex))
  {
    RequestManager::addSlice(slice);
    consensus = slice->getConsensus();
  }

  // Validate that the slice should be removed
  if (m_verifyEmpty)
  {
    if (removeCount && removeCount != sliceDepth)
    {
      cerr << "Not all reads were removed from slice at " 
           << origGindex << endl;
    }
    else if (!removeCount && consensus == '-' && sliceDepth)
    {
      const string & bases = slice->getNucData();

      if (bases.find_first_not_of('-') == bases.npos)
      {
        cerr << "Found an empty slice not reported at " 
             << origGindex << endl;
      }
    }
  }

  tcovType type = emptytype;

  // Print the slice if it hasn't be removed
  if (!removeCount || !sliceDepth)
  {
    type = edittype;

    long coordinate = m_newIndex;
    if (coordinate && consensus == '-') { coordinate--; }

    slice->resetIndex(coordinate, m_newGindex);
    slice->printXML(m_out);

    if (consensus != '-') { m_newIndex++; }
    m_newGindex++;
  }

  if (m_createExtendedTcov)
  {
    writeExtendedTcov(slice, type);
  }

  delete slice;
}

void SliceEditor::writeExtendedTcov(SliceData * slice, tcovType type)
{
  ofstream * out = m_origTcov;

  if (type == edittype || type == emptytype)
  {
    out = m_editTcov;
  }

  *out << slice->getGindex() << " "
      << slice->getIndex()  << " ";

  if (type != emptytype)
  {
    slice->setDoAmbiguity(0);
    libSlice::ConsensusData unresults = slice->doCalculations();

    slice->setDoAmbiguity(1);
    libSlice::ConsensusData ambresults = slice->doCalculations();

    *out << unresults.getConsensusBase() << " "
        << unresults.getQualityClass()  << " "

        << ambresults.getAmbiguityCode() << " "
        << ambresults.getQualityClass()  << " "

        << unresults.getQualityValue()  << " " // Both have the same cqv

        << slice->getNucData() << " ";

    int depth = slice->getDepth();

    int i;
    for (i = 0; i < depth; i++)
    {
      if (i) { *out << ":"; }
      *out << (int) slice->getQualVal(i);
    }
    
    *out << " ";

    for (i = 0; i < depth; i++)
    {
      if (i) { *out << ":"; }
      *out << slice->getRead(i)->getId();
    }

  }
  else
  {
    *out << "*";
  }

  *out << endl;
}


void SliceEditor::addReadData(ReadData * readData)
{
  int offset = 0;
  SeqEditMap::iterator edit;
  const string & seqname = readData->getSeqname();

  for (edit =  m_seqEdits.find(seqname);
       edit != m_seqEdits.end() && edit->first == seqname;
       edit++)
  {
    offset += edit->second->seqOffset();
  }

  if (m_verifyIndex) { m_seqPos[seqname] = readData->getSeqleft(); }

  if (offset)
  {
    // Adjust the end of the clear range to match

    long seqleft = readData->getSeqleft();
    long seqright = readData->getSeqright();

    if (readData->isReverseCompliment()) { seqleft += offset; }
    else                                 { seqright += offset; }

    char leftbuffer[32];
    char rightbuffer[32];

    snprintf(leftbuffer, sizeof(leftbuffer), "%ld", seqleft);
    snprintf(rightbuffer, sizeof(rightbuffer), "%ld", seqright);

    readData->addClrData(leftbuffer, rightbuffer);
  }

  readData->printXML(m_out);
}

void SliceEditor::addRequest(ConsensusRequest * request)
{
  m_requestData = request;

  printXMLHeader(request);

  m_out << "  <ReadCollection>" << endl;
}

void SliceEditor::writeExtendedTcovReadTable()
{
  list<ofstream *> tcovlist;
  list<ofstream *>::const_iterator i;

  tcovlist.push_back(m_origTcov);
  tcovlist.push_back(m_editTcov);

  for (i =  tcovlist.begin();
       i != tcovlist.end();
       i++)
  {
    **i << endl;
    **i << endl;

    ReadData * rd;
    m_readManager->resetReadIterator();
    while ((rd = m_readManager->getNextRead()) != NULL)
    {
      **i << rd->getId() << " "
          << rd->getSeqname() << " "
          << (int)rd->isReverseCompliment() << endl;
    }
  }
}

int SliceEditor::printResults()
{
  m_out << "  </SliceRange>" << endl;
  m_out << "</Request>"      << endl;

  if (m_createExtendedTcov) { writeExtendedTcovReadTable(); }

  return m_errorCount;
}

void SliceEditor::processEditFile(const string & filename)
{
  string contigId;
  long gindex;
  long index;
  long seqPos;
  string seqName;
  char origBase;
  char editBase;
  int rc;
  int qv;

  ifstream * editFile = FileSystem::openInputFile(filename);
  
  while (*editFile >> contigId 
                   >> gindex
                   >> index 
                   >> seqPos 
                   >> seqName 
                   >> origBase 
                   >> editBase
                   >> rc
                   >> qv)
  {
    Edit * newEdit = new Edit(contigId, gindex, index, 
                              seqPos, seqName, 
                              origBase, editBase, rc, qv,
                              m_tf->getLogger()->getCategory("Edit"),
                              m_editFilterLevel);
    
    //newEdit->printXML(m_out);

    m_indexEdits.insert(IndexEditPair(gindex, newEdit));
    m_seqEdits.insert(SeqEditPair(seqName, newEdit));
  }

  delete editFile;
}

void SliceEditor::setVerifyIndex(int verifyIndex)
{
  m_verifyIndex = verifyIndex;
}

void SliceEditor::setVerifyEmpty(int verifyEmpty)
{
  m_verifyEmpty = verifyEmpty;
}

void SliceEditor::setCreateExtendedTcov(int createExtendedTcov)
{
  m_createExtendedTcov = createExtendedTcov;
}

void SliceEditor::handleOptions(Foundation * tf)
{
  tf->getOptions()->addOptionResult("verify-index|i", &m_verifyIndex,
                                    "verify ungapped index");

  tf->getOptions()->addOptionResult("verify-empty|v", &m_verifyEmpty,
                                    "verify empty slices");

  tf->getOptions()->addOptionResult("editFilterLevel|f=i", &m_editFilterLevel,
                                    "Set the filter level for edits");

  tf->getOptions()->addOptionResult("output-prefix|p=s", &m_outfilePrefix,
                                    "Specify the prefix for the tcov files");

  RequestManager::handleOptions(tf);

  if (!m_doConsensus &&
      !m_doAmbiguity &&
      !m_verifyIndex)
  {
    m_doConsensus   = 1;
    m_doAmbiguity   = 1;
    m_verifyIndex   = 1;
    m_verifyEmpty   = 1;
  }

  if (m_outfilePrefix != "")
  {
    m_createExtendedTcov = 1;

    m_origTcov = RequestManager::openOutputFile(m_outfilePrefix+".orig.etcov");
    m_editTcov = RequestManager::openOutputFile(m_outfilePrefix+".edit.etcov");
  }
}
