#include "SliceData.hh"
#include <stdlib.h>  // for atol

/*! @file
 *  @brief Defines implementation for SliceData
 */

namespace libSlice
{
  using std::string;

//! Constructor initializes the slice information
/*! The read manager needs to be provided so that reverse compliment
 *  flags can be looked up.
 *
 *  @param readManager Manager of all associated reads
 *  @param index Index of slice [optional]
 *  @param gindex Gindex of slice [optional]
 *  @param existing Existing consensus of slice [optional default=' ']
 *  @param sliceRangeIndex Ungapped Index of the parent SliceRange
 *  @param sliceRangeGindex Gapped index of the parent SliceRange
 */
SliceData::SliceData(const ReadManager & readManager,
                     const char * index, 
                     const char * gindex,
                     const char * existing,
                     long sliceRangeIndex,
                     long sliceRangeGindex)
 : m_readManager(readManager)
{
  m_slice = NULL;
  m_highQualityThreshold = 0;

  m_nucData     = "";
  m_readIdData  = "";
  m_qualValData = "";

  m_errorData = "";
  m_errorCode = SLICE_NOERROR;
  m_errorOperation = STROP_NONE;

  m_doAmbiguity = 0;
  m_scoringModel = 0;

  m_index    = 0;
  m_gindex   = 0;
  m_indexstr = "Unknown";
  m_existing = (existing) ? existing :  "-";

  if (index)
  {
    m_index = atol(index);
    m_index += sliceRangeIndex;

    m_indexstr = index;
  }

  if (gindex)
  {
    m_gindex = atol(gindex);
    m_gindex += sliceRangeGindex;
  }


  double normal = 0.2;
  setDistributionTable(normal, normal, normal, normal, normal);
}


//! Deletes the flattened slice
/*! If the slice has been flattened into a struct slice, deletes the pointer
 *  to that struct and the data fields that are pointers.
 */
SliceData::~SliceData()
{
  if (m_slice)
  {
    if (m_slice->rc) delete [] m_slice->rc;
    if (m_slice->qv) delete [] m_slice->qv;
    if (m_slice->bc) delete [] m_slice->bc;

    delete m_slice;
    m_slice = NULL;
  }
}

//! Sets the distribution table
/*! This is the table used by getConsQV
 */
void SliceData::setDistributionTable(double freqA, double freqC,
                                     double freqG, double freqT,
                                     double freqGap)
{
  m_distribution.freqA   = freqA;
  m_distribution.freqC   = freqC;
  m_distribution.freqG   = freqG;
  m_distribution.freqT   = freqT;
  m_distribution.freqGap = freqGap;
}

//! Sets the threshold of high quality for getConsensus and getConsQC
void SliceData::setHighQualityThreshold(int highQualityThreshold)
{
  m_highQualityThreshold = highQualityThreshold;
}

//! Appends data to the list of Nuc Data
/*! Useful because Xerces won't guarantee that all of the character data will
 *  come in a single message, which is how Nuc data is specified.
 */
void SliceData::addNuc(const char * nucData)
{
  if (nucData) m_nucData.append(nucData);
}

//! Appends data to the list of QualVal Data
/*! Useful because Xerces won't guarantee that all of the character data will
 *  come in a single message, which is how QualVal data is specified.
 */
void SliceData::addQualVal(const char * qualVal)
{
  if (qualVal) m_qualValData.append(qualVal);
}

//! Appends data to the list of ReadId Data
/*! Useful because Xerces won't guarantee that all of the character data will
 *  come in a single message, which is how ReadId data is specified
 */
void SliceData::addReadId(const char * readId)
{
  if (readId) m_readIdData.append(readId);
}

//! Force the consensus to be a certain value
void SliceData::setConsensus(char consensus)
{
  m_existing = consensus;
  if (m_slice)
  {
    m_slice->c = consensus;

  }
}

//! Splits a string into elements and runs a calculation on each element
/*! Handy utility function that is roughly equivalant to foreach in perl
 *  
 *  @param data Reference to the string to split
 *  @numresults Number of results required
 *  @param operation Which operation to perform
 *  @param underflowError Message to throw when there is less data than expected
 *  @param overflowError Message to throw when there is more data than expected
 * 
 *  @returns A c string of calculation results numresults long
 *  @throws Exception If there is more or less data than required
 */
// TODO: Investigate strtok
char * SliceData::foreach(const string & data, int numresults,
                          StringOperation operation,
                          const char * underflowError,
                          const char * overflowError)
{
  const char * databuf = data.c_str();
  char * retval = new char [numresults+1];
  retval[numresults] = '\0';

  string::size_type startpos = 0;
  string::size_type endpos = 0;

  for (int i=0; i < numresults; i++)
  {
    endpos = data.find(' ', startpos);

    if (endpos == data.npos)
    {
      if (i != (numresults-1))
      {
        m_errorOperation = operation;
        m_errorCode = SLICE_ERR_UNDERFLOW;

        throw Exception(underflowError, "SliceData::foreach");
      }

      endpos = data.length() + 1;
    }
    
    char elementResult = '\0';

    if (operation == STROP_GET_QUAL_VAL)
    {
      elementResult = (char) (atoi(databuf + startpos) & 0xff);
    }
    else if (operation == STROP_IS_REVERSE_COMPLIMENT)
    {
      const string & element = data.substr(startpos, (endpos - startpos));
      const ReadData * rd = m_readManager.find(element.c_str());
      
      if (!rd)
      {
        m_errorOperation = operation;
        m_errorCode = SLICE_ERR_NO_READID;
        m_errorData = element;

        throw Exception("ReadID " + element + " not found",
                            "SliceData::foreach");
      }

      elementResult = rd->isReverseCompliment();
    }

    retval[i] = elementResult;
    startpos = endpos + 1;
  }

  if (numresults && (endpos != data.length()+1))
  {
    m_errorOperation = operation;
    m_errorCode = SLICE_ERR_OVERFLOW;

    throw Exception(overflowError, "SliceData::foreach");
  }

  return retval;
}

//! Transforms the raw buffer of quality values into an array
/*! Preprocess the data to verify that all of the quality values are numberic.
 *
 *  @param len Required number of quality values to extract
 *  @throws Exception If there is non-numeric data in the buffer.
 *  @returns An arrray of characters, each is the typecasted quality value
 *  @see foreach
 */
char * SliceData::getQualVals(int len)
{
  // There better only be number data in m_qualValData
  if (m_qualValData.find_first_not_of("0123456789 ") != m_qualValData.npos)
  {
    m_errorCode = SLICE_ERR_QV_NONNUMERIC;
    m_errorOperation = STROP_GET_QUAL_VAL;

    throw Exception("Invalid Quality Value in slice " + m_indexstr,
                        "SliceData::getQualVals");
  }

  return foreach(m_qualValData, len, STROP_GET_QUAL_VAL,
                 "Number of Bases > Number of QualVals",
                 "Number of QualVals > Number of Bases");
}

//! Transforms the raw buffer of read ids into an array of reverse compliments
/*! Calls foreach with requested the data to be extracted and lookups the 
 *  reverse compliments for each read id and stores the results into an array
 *
 *  @param len Required number of reverse compliments to extract
 *  @returns An arrray of characters, each is the typecasted reverse compliments
 *  @see foreach
 */
char * SliceData::getReverseCompliments(int len)
{
  return foreach(m_readIdData, len, STROP_IS_REVERSE_COMPLIMENT,
                 "Number of Bases > Number of ReadIds",
                 "Number of ReadIds > Number of Bases");
}

//! Gets the existing consensus
/*! Returns the existing consensus. If the existing consensus was not a single
 *  characters, returns '-'
 *  
 */
char SliceData::getExisting() const
{
  char retval = '-';

  if (m_existing.length() == 1)
  {
    retval = m_existing[0];
  }
  else
  {
    //TODO: Decide what validation should be done (if any)
  }

  return retval;
}

//! Returns the depth of coverage based on the number of bases specified.
unsigned short SliceData::getDepth() const
{
  return m_nucData.length();
}

//! Copies the bases into a new c-string of bases
char * SliceData::getBases(int len) const
{
  char * retval = new char [len + 1];
  strcpy(retval, m_nucData.c_str());
  return retval;
}

//! Flattens the SliceData object information into a struct slice
/*! Fills out the data of a struct slice from the SliceData information.
 *
 *  @see getDepth
 *  @see getExisting
 *  @see getBases
 *  @see getQualVals
 *  @see getReverseCompliments
 */
void SliceData::flattenSlice() 
{
  if (!m_slice)
  {
    m_slice = new libSlice_Slice;
    memset (m_slice, 0, sizeof(libSlice_Slice));

    int depth = getDepth();

    m_slice->dcov = depth;
    m_slice->c    = getExisting();
    m_slice->bc   = getBases(depth);
    m_slice->qv   = getQualVals(depth);
    m_slice->rc   = getReverseCompliments(depth);
  }
}

//! Gets bc[index]
char SliceData::getBase(unsigned short index)
{
  flattenSlice();

  if (index > getDepth())
  {
    throw(Exception("Index > Depth", "getBase"));
  }

  return m_slice->bc[index];
}

//! Gets qv[index]
char SliceData::getQualVal(unsigned short index)
{
  flattenSlice();

  if (index > getDepth())
  {
    throw(Exception("Index > Depth", "getQualVal"));
  }

  return m_slice->qv[index];
}

//! Gets ReadId[index]
//TODO: Refactor to utilize foreach
ReadData * SliceData::getRead(unsigned short index) const
{
  if (index > getDepth())
  {
    throw(Exception("Index > Depth", "getRead"));
  }

  string::size_type startpos = 0;
  string::size_type endpos = 0;

  while (index)
  {
    startpos = m_readIdData.find(' ', startpos);
    startpos++;

    if (startpos == m_readIdData.npos)
    {
      throw Exception("Index > Number of ReadIds", "getReadId");
    }

    index--;
  }

  endpos = m_readIdData.find(' ', startpos);
  if (endpos == m_readIdData.npos)
  {
    endpos = m_readIdData.length();
  }

  const string & element = m_readIdData.substr(startpos, 
                                                    (endpos - startpos));
  ReadData * rd = m_readManager.find(element.c_str());
      
  if (!rd)
  {
    throw Exception("ReadID " + element + " not found",
                        "SliceData::getRead");
  }

  return rd;
}

/*
 * This turned out to be slower than getRead
 * 
const ReadData * SliceData::getRead2(unsigned short index)
{
  int depth = getDepth();

  if (index > depth)
  {
    throw(Exception("Index > Depth", "getRead2"));
  }

  if (m_readVector.empty()) 
  {
    if (depth > 1)
    {
      int len = m_readIdData.length();
      char buffer[len+1];
      strcpy(buffer, m_readIdData.c_str());

      char * start = NULL;

      for (int i = 0; i < len; i++)
      {
        if (buffer[i] != ' ' && start == NULL)
        {
          start = &buffer[i];
        }
        else if (buffer[i] == ' ')
        {
          buffer[i] = '\0';

          if (start)
          {
            // Start points to the beginning of the null terminated string to
            // extract
            
            m_readVector.push_back(m_readManager.find(start));
            start = NULL;
          }
        }
      }
    }
    else
    {
      m_readVector.push_back(m_readManager.find(m_readIdData.c_str()));
    }

    if (m_readVector.size() != depth)
    {
      throw Exception("Number of reads doesn't match number of bases",
                          "SliceData::getRead2");

    }
  }

  return m_readVector[index];
}

*/

//! Returns the ungapped index of the slice
/*! The ungapped index is index + sliceRangeIndex
 */
long SliceData::getIndex() const
{
  return m_index;
}

//! Returns the gapped index of the slice
/*! The gapped index is gindex + sliceRangeGindex
 */
long SliceData::getGindex() const
{
  return m_gindex;
}

//! Gets the existing consensus
char SliceData::getConsensus()
{
  flattenSlice();

  return m_slice->c;
}


//! Flattens a slice, and calls libSlice to do the quality calculations
/*! A new consensus is always calculated before the quality class or quality
 *  value is calculated.
 *
 *  The new consensus is set to be used as the "Existing" attribute of the
 *  XML slice
 *
 *  @see flattenSlice()
 */
ConsensusData SliceData::doCalculations(bool recallEmpty)
{
  flattenSlice();
  int errCode;

  libSlice_Consensus resultConsensus;

  // This should probably never be set
  if (recallEmpty) { libSlice_setRecallEmpty(recallEmpty); }


  // Don't do ambiguity here, get that separately
  if ((errCode = libSlice_getConsensusParam(m_slice, 
                                            &resultConsensus, 
                                            &m_distribution, 
                                            m_highQualityThreshold,
                                            0)))
  {
    m_errorCode = SLICE_ERR_GETCONSENSUS;

    throw Exception("Error Calculating consensus",
                        "_getConsensus");
  }
  
  if (m_doAmbiguity == 2)
  {
    char origFlags = resultConsensus.ambiguityFlags;
    libSlice_updateAmbiguity(m_slice, &resultConsensus, m_highQualityThreshold);

    if (0 && origFlags != resultConsensus.ambiguityFlags)
    {
      cerr << "Annotation: Updating Ambiguity " 
           << libSlice_convertAmbiguityFlags(origFlags)
           << " to " 
           << libSlice_convertAmbiguityFlags(resultConsensus.ambiguityFlags)
           << " at gindex " << m_gindex << endl;
    }
  }
  else if (m_doAmbiguity == 3)
  {
    char origFlags = resultConsensus.ambiguityFlags;
    libSlice_updateAmbiguityConic(m_slice, 
                                  &resultConsensus, 
                                  m_highQualityThreshold);
                                  
    if (0 && origFlags != resultConsensus.ambiguityFlags)
    {
      cerr << "Conic: Updating Ambiguity " 
           << libSlice_convertAmbiguityFlags(origFlags)
           << " to " 
           << libSlice_convertAmbiguityFlags(resultConsensus.ambiguityFlags)
           << " at gindex " << m_gindex << endl;
    }
  }

  char ambiguityCode;

  if (m_slice->dcov == 0 && !recallEmpty)
  {
    // Don't update the slice
    ambiguityCode = m_slice->c;
  }
  else
  {
    ambiguityCode = libSlice_convertAmbiguityFlags(resultConsensus.ambiguityFlags);
    if (m_doAmbiguity)
    {
      m_slice->c = ambiguityCode;
    }
    else
    {
      m_slice->c = resultConsensus.consensus;
    }
  }

  m_existing = (string) "" + m_slice->c;

  char qualityClass = libSlice_getConsQC(m_slice, m_highQualityThreshold);

  return ConsensusData(&resultConsensus, 
                       ambiguityCode, 
                       qualityClass, 
                       m_doAmbiguity);
}

//! Prints a slice to a specified output stream
/*! Flattens the slice first so that the triplet of (base, qv, rc) can be
 *  printed for each base. Data is enclosed in \<Slice ...\> ... \</Slice\>
 *  tags and each base is written as \<Base ...\>
 *
 *  This does not produce valid Slice XML, but shows how the data was actually
 *  parsed.
 *  
 *  @see flattenSlice
 */
void SliceData::print(ostream & out, const string & prefix)
{
  flattenSlice();

  out << prefix
      << "<Slice Index=\"" << m_index
      << "\" Gindex=\""    << m_gindex
      << "\" Existing=\""  << m_existing
      << "\" Depth=\""     << getDepth()
      << "\">" << endl;

  for (int i=0; i < m_slice->dcov; i++)
  {
    out << prefix
        << "  <Base i=\"" << i
        << "\" bc=\"" <<       m_slice->bc[i] 
        << "\" qv=\"" << (int) m_slice->qv[i]
        << "\" rc=\"" << (int) m_slice->rc[i]
        << "\" />" << endl;
  }

  out << prefix << "</Slice>" << endl;
}

//! Prints a slice to a specified output stream in Slice XML
/*! This does produce valid XML. The bases, qualvals, and readids come from
 *  the unparsed add* functions.
 *
 *  The slice is not flattened
 */
void SliceData::printXML(ostream & out, const string & prefix)
{
  out << prefix
      << "<Slice Index=\"" << m_index
      << "\" Gindex=\""    << m_gindex
      << "\" Existing=\""  << m_existing
      << "\">" << endl;

  out << prefix << "  <Nuc>"     << m_nucData     << "</Nuc>"     << endl;
  out << prefix << "  <Qualval>" << m_qualValData << "</Qualval>" << endl;
  out << prefix << "  <ReadID>"  << m_readIdData  << "</ReadID>"  << endl;

  out << prefix << "</Slice>" << endl;
}

//! Resets the read id data
/*! Truncates the data added with addReadId()
 *
 *  Useful so a slice can be modified without making a copy.
 *
 *  Note: Any subsequent attempts to flatten the slice for calculations will
 *  potentially fail if the readId values are changed to ones that do not occur
 *  in the ReadManager
 */
void SliceData::resetReadId()
{
  m_readIdData = "";
  
  if (m_slice)
  {
    delete m_slice;
    m_slice = NULL;
  }
}

//! Resets the QualVal data
/*! Truncates the data added with addQualVal()
 *
 *  Useful so a slice can be modified without making a copy
 */
void SliceData::resetQualVal()
{
  m_qualValData = "";

  if (m_slice)
  {
    delete m_slice;
    m_slice = NULL;
  }
}

//! Resets the Nuc data
/*! Truncates the data added with addNuc()
 *
 *  Useful so a slice can be modified without making a copy
 */
void SliceData::resetNuc()
{
  m_nucData = "";

  if (m_slice)
  {
    delete m_slice;
    m_slice = NULL;
  }
}

//! Resets the index
/*! Overwrites the index and gindex provided at construction
 *
 *  Useful so a slice can be modified without making a copy
 */
void SliceData::resetIndex(long index, long gindex)
{
  m_index = index;
  m_gindex = gindex;
}

//! Gets the complete qualval string set by addQualVal
const string & SliceData::getQualValData() const
{
  return m_qualValData;
}

//! Gets the complete readid string set by addReadId
const string & SliceData::getReadIdData() const
{
  return m_readIdData;
}

//! Gets the complete nuc string set by addNuc
const string & SliceData::getNucData() const
{
  return m_nucData;
}

//! Prints out the error information from the slice to a given output stream
/*! Before an exception is thrown from SliceData, information is stored in
 *  the slice so that the details of the error can be displayed along with
 *  all of the other information (if requested)
 */
void SliceData::printError(ostream & out) const
{
  if (m_errorCode != SLICE_NOERROR)
  {
    out << "  Error Slice index=" << m_index
        << " code=" << m_errorCode
        << " msg=";

    switch (m_errorCode)
    {
      case SLICE_NOERROR:
         out << "No Error";
         break;

      case SLICE_ERR_QV_NONNUMERIC:
         out << "Non Numberic values in Quality Value";
         break;

      case SLICE_ERR_UNDERFLOW:
         out << "Less data than required in operation";
         break;

      case SLICE_ERR_OVERFLOW:
         out << "More data than required in operation";
         break;

      case SLICE_ERR_NO_READID:
         out << "ReadId not found"
             << " readid=" << m_errorData;
         break;

      case SLICE_ERR_GETCONSENSUS:
         out << "Internal error in getConsensus";
         break;

      case SLICE_ERR_CALCULATION_NOT_DONE:
         out << "Results requested before calculations performed";
         break;
    };

    if (m_errorCode == SLICE_ERR_OVERFLOW || m_errorCode == SLICE_ERR_UNDERFLOW)
    {
      switch (m_errorOperation)
      {
        case STROP_GET_QUAL_VAL:
          out << " operation=" << m_errorOperation
              << " op_str="    << "getQualval";
          break;

        case STROP_IS_REVERSE_COMPLIMENT:
          out << " operation=" << m_errorOperation
              << " op_str="    << "getReverseCompliment";
          break;

        case STROP_NONE:
          break;
      };
    }
   
//    out << endl;
  }
}

//! Gets the internal error code that threw an exception
int SliceData::getErrorCode() const
{
  return m_errorCode;
}

//! Gets the code for the string operation threw the exception
int SliceData::getErrorOperation() const
{
  return m_errorOperation;
}

//! Set if consensus and quality class should be based on an ambiguity code
/*! 0 means no ambiguity<br>
 *  1 means do minimal Churchill&Waterman<br>
 *  2 means do minimal Churchill&Waterman and updateAmbiguity<br>
 *  3 means do minimal Churchill&Waterman and updateAmbiguityConic<br>
 */
void SliceData::setDoAmbiguity(int doAmbiguity)
{
  m_doAmbiguity = doAmbiguity;
}

//! Get the compliment of a base or ambiguity code
char SliceData::getBaseCompliment(char base)
{
  return libSlice_getCompliment(base);
}

void SliceData::setScoringModel(int scoringModel)
{
  m_scoringModel = scoringModel;
}

}

//! Convienent way to print slices out
/*! 
 * Note this prints a slice, not printXML a slice
 */
ostream & operator<< (ostream & os, libSlice::SliceData & slice)
{
  slice.print(os);
  return os;
}

// Convienent way to print *slices out
/*! 
 * Note this prints a slice, not printXML a slice
 */
ostream & operator<< (ostream & os, libSlice::SliceData * slice)
{
  slice->print(os);
  return os;
}
